from unittest import TestCase


class GithubApiTests(TestCase):
    def test_clone_github_repository(self):
        pass

    def test_create_feature_branch(self):
        pass

    def test_run_script_inside_cloned_repository(self):
        pass

    def test_commit_changes_made_to_cloned_repository(self):
        pass

    def test_push_commits(self):
        pass

    def test_create_pull_request(self):
        pass

    def test_execution_logs(self):
        # Test that pull request link is present in the logs
        # Test that error Tracebacks are in the logs
        pass



import os.path
from typing import Dict, Optional
from urllib.parse import urlparse

import git
from git.cmd import Git

from gitmultirepoupdater.data_types import RepoState
from gitmultirepoupdater.constants import CloningStates
from gitmultirepoupdater.utils.throttled_tasks_executor import ThrottledTasksExecutor
from gitmultirepoupdater.utils.helpers import get_access_token, get_default_branch


def get_repo_access_url(url: str) -> Optional[str]:
    """Converts repository url to url which is suitable for cloning"""

    if access_token := get_access_token(url):
        parsed_url = urlparse(url)
        domain_with_access_token = f"api:{access_token}@{parsed_url.netloc.split('@')[-1]}"
        parsed_url = parsed_url._replace(netloc=domain_with_access_token, scheme="https")
        return parsed_url.geturl()
    return None


async def clone_repository(repo: RepoState) -> None:
    """Clones repository with default (or source) branch."""

    clone_to = repo.args.clone_to
    repo.directory = os.path.join(clone_to, repo.name)

    # TODO: add a way to clone using access token: https://stackoverflow.com/questions/25409700/using-gitlab-token-to-clone-without-authentication/29570677#29570677
    # git clone https://:YOURKEY@your.gilab.company.org/group/project.git

    # TODO: add ssh support: urls like git@gitlab.com:niekas/gitlab-api-tests.git

    try:
        if os.path.exists(repo.directory):  # If repository exists: clean it, pull changes, checkout default branch
            g = Git(repo.directory)
            g.clean("-dfx")
            g.execute(["git", "fetch", "--all"])

            default_branch = get_default_branch(repo)
            g.checkout(default_branch)

            repo.cloning_state = CloningStates.CLONED.value
        else:
            if repo_access_url := get_repo_access_url(repo.url):
                git.Repo.clone_from(repo_access_url, repo.directory)

                g = Git(repo.directory)
                g.execute(["git", "fetch", "--all"])

                repo.cloning_state = CloningStates.CLONED.value
            else:
                repo.cloning_state = CloningStates.ACCESS_TOKEN_NOT_PROVIDED.value

        repo.target_branch = repo.target_branch or get_default_branch(repo)

    except git.exc.GitCommandError:
        repo.cloning_state = CloningStates.NOT_FOUND.value


def print_cloned_repositories(repos):
    print()
    print("\033[1;34m|" + "Cloned repositories".center(77, "-") + "|\033[0m")
    should_print_not_cloned_repos = False
    for repo in repos.values():
        if repo.cloning_state == CloningStates.CLONED.value:
            print(f"\033[1;34m|\033[0m - {repo.url.ljust(73, ' ')} \033[1;34m|\033[0m")
        else:
            should_print_not_cloned_repos = True
    if should_print_not_cloned_repos:
        print("\033[1;34m|\033[0m" + "Did NOT clone these repositories:".center(77, "-") + "\033[1;34m|\033[0m")
        for repo in repos.values():
            if repo.cloning_state != repo.cloning_state:
                print(f"\033[1;34m|\033[0m - {(repo.url + ' ' + CloningStates.CLONED.value).ljust(73, ' ')} \033[1;34m|\033[0m")
    print("\033[1;34m|" + "".center(77, "-") + "|\033[0m")


def clone_repositories(repos: Dict[str, RepoState], executor: ThrottledTasksExecutor) -> None:
    for repo in repos.values():
        executor.run(clone_repository(repo))
    executor.wait_for_tasks_to_finish()
    print_cloned_repositories(repos)



import logging

import os
import os.path
from typing import Dict, List

from gitmultirepoupdater.data_types import CliArguments, RepoState
from gitmultirepoupdater.utils.helpers import get_repo_name, get_repo_owner, get_domain, to_kebab_case

logger = logging.getLogger()


def is_url_or_git(file_names_or_repo_url: str) -> bool:
    # TODO: use urlparse to verify if its url and use regexp for git url
    return ".com" in file_names_or_repo_url.lower()


def read_repositories_from_file(repos_filename) -> List[str]:
    """Reads a list of repositories from a file while ignoring commented out lines."""
    with open(repos_filename) as f:
        return [l.strip() for l in f.readlines() if not l.strip().startswith("#")]


def get_repository_states(args: CliArguments) -> Dict[str, RepoState]:
    repo_urls = []
    for file_names_or_repo_url in args.repos:
        if not is_url_or_git(file_names_or_repo_url) and os.path.exists(file_names_or_repo_url):
            newly_read_repos = read_repositories_from_file(file_names_or_repo_url)
            repo_urls.extend(newly_read_repos)
        else:
            repo_urls.append(file_names_or_repo_url)


    branch = args.branch or to_kebab_case(args.commit_message)

    repos: Dict[str, RepoState] = {}
    for repo_url in repo_urls:
        repo_name = get_repo_name(repo_url)
        repo_owner = get_repo_owner(repo_url)
        domain = get_domain(repo_url)

        repos[repo_name] = RepoState(
            args=args,
            name=repo_name,
            owner=repo_owner,
            url=repo_url,
            domain=domain,
            branch=branch,
        )

    return repos



import json
from logging import getLogger
from typing import Dict

import httpx
from gitmultirepoupdater.constants import PullRequestStates

from gitmultirepoupdater.data_types import RepoState, HttpRequestParams
from gitmultirepoupdater.utils.helpers import get_access_token
from gitmultirepoupdater.utils.throttled_tasks_executor import ThrottledTasksExecutor

logger = getLogger()


def get_http_request_params_for_pull_request_creation(repo: RepoState) -> HttpRequestParams:
    """
    Gitlab create MR docs: https://docs.gitlab.com/ee/api/merge_requests.html#create-mr
    Github create MR docs: https://docs.github.com/en/rest/pulls/pulls?apiVersion=2022-11-28#create-a-pull-request
    """
    if repo.domain == "github.com":
        url = f"https://api.github.com/repos/{repo.owner}/{repo.name}/pulls"
        headers = {
          "Accept": "application/vnd.github+json",
          "Authorization": f"Bearer {get_access_token(repo.url)}",
          "X-GitHub-Api-Version": "2022-11-28"
        }
        data = {
            "title": repo.args.commit_message,
            "body": repo.args.commit_message,
            "head": repo.branch,
            "base": repo.target_branch,
        }

    else:  # Use gitlab.com API by default
        url = f"https://{repo.domain}/api/v4/projects/{repo.owner}%2F{repo.name}/merge_requests"
        headers = {
            "PRIVATE-TOKEN": get_access_token(repo.url)
        }
        data = {
            "source_branch": repo.branch,
           "target_branch": repo.target_branch,
            "title": repo.args.commit_message,
        }
    return HttpRequestParams(
        url=url,
        headers=headers,
        data=data,
    )


def print_pull_requests(repos):
    print()
    print("\033[1;34m|" + "Created Pull Requests".center(77, "-") + "|\033[0m")
    show_not_created_pull_requests = False
    for repo in repos.values():
        if repo.pull_request_state == PullRequestStates.CREATED.value:
            print(f"\033[1;34m|\033[0m - {repo.pull_request_url.ljust(73, ' ')} \033[1;34m|\033[0m")
        else:
            show_not_created_pull_requests = True
    if show_not_created_pull_requests:
        print("\033[1;34m|" + "Not created Pull Requests".center(77, "-") + "|\033[0m")
        for repo in repos.values():
            if repo.pull_request_state == PullRequestStates.GOT_BAD_RESPONSE.value:
                print(f"\033[1;34m|\033[0m - {repo.url.ljust(73, ' ')} \033[1;34m|\033[0m")
                print(f"\033[1;34m|\033[0m   status_code={repo.pull_request_status_code} reason={repo.pull_request_reason} \033[1;34m|\033[0m")
    print("\033[1;34m|" + "".center(77, "-") + "|\033[0m")


async def create_pull_request(repo: RepoState):
    # https://stackoverflow.com/questions/56027634/creating-a-pull-request-using-the-api-of-github
    request_params = get_http_request_params_for_pull_request_creation(repo)

    async with httpx.AsyncClient() as client:
        response = await client.post(
            url=request_params.url,
            headers=request_params.headers,
            data=request_params.data,
        )
        if response.status_code < 400:
            repo.pull_request_state = PullRequestStates.CREATED.value
            repo.pull_request_url = response.json().get("web_url", response.json().get("url"))
        else:
            repo.pull_request_state = PullRequestStates.GOT_BAD_RESPONSE.value
            repo.pull_request_status_code = response.status_code
            repo.pull_request_reason = json.dumps(response.json())


def create_pull_request_for_each_repo(repos: Dict[str, RepoState], executor: ThrottledTasksExecutor) -> None:
    for repo in repos.values():
        executor.run(create_pull_request(repo))
    executor.wait_for_tasks_to_finish()
    print_pull_requests(repos)



from enum import Enum


ACCESS_TOKEN_VAR_NAMES = {
    "gitlab.com": "GITLAB_ACCESS_TOKEN",
    "github.com": "GITHUB_OAUTH_TOKEN",
    "DEFAULT": "GIT_TOKEN",
}


class CloningStates(Enum):
    NOT_STARTED = "NOT_STARTED"
    ACCESS_TOKEN_NOT_PROVIDED = "ACCESS_TOKEN_NOT_PROVIDED"
    CLONING = "CLONING"
    CLONED = "CLONED"
    NOT_FOUND = "NOT_FOUND"
    SKIPPED = "SKIPPED"


class PullRequestStates(Enum):
    NOT_CREATED = "NOT_CREATED"
    CREATED = "CREATED"
    GOT_BAD_RESPONSE = "GOT_BAD_RESPONSE"
    MERGED = "MERGED"


class ModificationState(Enum):
    NOT_STARTED = "NOT_STARTED"
    GOT_EXCEPTION = "GOT_EXCEPTION"
    MODIFIED = "MODIFIED"
    NO_FILES_CHANGED = "NO_FILES_CHANGED"
    PUSHED_TO_REMOTE = "PUSHED_TO_REMOTE"



# import asyncio
# from typing import Callable, Coroutine, Optional
# from threading import Thread


#####################
# Reactor prototype #
#####################


# ##############
# # helpers.py #
# ##############

# # What are the limitations of this class?


# class ThrottledTasksExecutor:
#     def __init__(self, delay_between_tasks: float = 0.2, in_separate_process: bool = False):
#         self.loop = asyncio.new_event_loop()
#         self.running_tasks = set()  # TODO: Find the example with task removal from the set.
#                                     # TODO: Would be nice to have awaitable future, instead of infinite loop.
#         self.delay_between_tasks = 1
#         self.in_separate_process = in_separate_process

#         self.can_task_be_executed = asyncio.Condition(loop=self.loop)
#         self.is_running = False

#     def start(self, in_separate_process: Optional[bool] = None):
#         """Starts a thread (or a process), which executes coroutines provided to the ThrottledTasksExecutor"""
#         if in_separate_process is None:
#             in_separate_process = self.in_separate_process
#         # TODO: investigate a way to start coroutines in a separate process:
#         #   - https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
#         # Param: in_separate_process: bool = False
#         thread = Thread(target=self._run_event_loop, daemon=True)
#         thread.start()

#         self.is_running = True

#         # Periodically emit permission to execute one task
#         asyncio.run_coroutine_threadsafe(self.allow_task_execution(every=self.delay_between_tasks), self.loop)

#     def stop(self):
#         """Terminates a thread (or a process), which executes coroutines provided to the ThrottledTasksExecutor"""
#         self.loop.stop()
#         self.is_running = False

#     def run(self, coroutine: Coroutine, callback: Callable):
#         """Executes coroutine in an executor thread, which makes sure not to hit throttling limits."""
#         # TODO: could add support for coroutines:
#         # if func is not coroutine and func is callable:
#         #     coroutine = func(*args, **kwargs)
#         #     if coroutine is not coroutine:
#         #         raise ValueError("Could not find coroutine")

#         # async with can_make_api_call:
#         #     task_result = task.result()

#         task = asyncio.run_coroutine_threadsafe(self.throttled_task(coroutine), self.loop)
#         self.running_tasks.add(task)
#         task.add_done_callback(self.mark_task_done(callback))

#     async def allow_task_execution(self, every: float, count: int = 1):
#         """Periodically emits event, which allows for one task to be executed."""
#         print("Started emitting execution events.")
#         while self.is_running:
#             print("Allowing one task execution.")
#             async with self.can_task_be_executed:   # TODO: is this async context really needed?
#                 self.can_task_be_executed.notify(n=count)
#             print("Allowed one task execution.")
#             await asyncio.sleep(every)
#         print("Stopped emitting execution events.")

#     def throttled_task(self, coroutine: Coroutine) -> Coroutine:
#         async def throttled_task_wrapper(*args, **kwargs):
#             async with self.can_task_be_executed:   # TODO: is this async context really needed?
#                 await self.can_task_be_executed.wait()
#             task_result = await coroutine
#             return task_result # task_result
#         return throttled_task_wrapper()

#     def mark_task_done(self, callback):
#         def task_done_wrapper(task):
#             task_result = task.result()
#             callback_result = callback(task_result)
#             self.running_tasks.discard(task)
#             return callback_result
#         return task_done_wrapper

#     def __enter__(self) -> "ThrottledTasksExecutor":
#         self.start()
#         return self

#     def __exit__(self, *exc_info):
#         asyncio.run(self.wait_for_tasks_to_finish())
#         self.stop()

#     async def wait_for_tasks_to_finish(self):
#         print(f"Started waiting for {len(self.running_tasks)} tasks to finish")
#         while self.running_tasks:
#             print(f"  Still waiting for {len(self.running_tasks)} tasks to finish")
#             await asyncio.sleep(0.3)

#     def _run_event_loop(self):
#         asyncio.set_event_loop(self.loop)
#         self.loop.run_forever()





# ## Plan:
# # Define task, which waits some time. And prints something.
# # Define function, which puts the tasks to be executed (multiple of them). 
#     # Test throtling: with diagram.
# # Execute some load testing on those functions.



# async def make_api_call(api_call_params) -> str:
#     # TODO: Can the task be a synchronous function?
#     # async with can_make_api_call:  # TODO: add this Condition
#     #     await can_make_api_call.wait()
#     print(f"{api_call_params} Making API call..")
#     await asyncio.sleep(2.4)
#     print(f"{api_call_params} Made API call..")
#     return str(api_call_params)


# def process_response(response: str):
#     print(f"Saving HTTP response with status code: {response}")
#     print(f"Finished processing the HTTP response {response}.") 
#     print(f"Removing the task from the pool of running tasks")


# ## TODO:
# # Implement that Condition.
# # Make real API calls using httpx.
# # Generalise the executor for different types of tasks: git clone and API calls.




# with ThrottledTasksExecutor(delay_between_tasks=0.2) as executor:
#     print("Creating the tasks:")
#     for api_call_params in range(10):
#         print(f"{api_call_params} Created task")
#         # TODO: I can put some arguments to the callback with functools.partial()
#         executor.run(make_api_call(api_call_params), callback=process_response)


#     # task = asyncio.run_coroutine_threadsafe(make_api_call(api_call_params), loop)
#     # running_tasks.add(task)
#     # print(f"Created task with id {api_call_params}")

#     # task.add_done_callback(process_http_response)


# #############
# ## Helpers ##
# #############





# print("All tasks have finished, exiting...")


# ## Problem:
# # Gather the dynamic set of tasks.
# # Should wait till all of the tasks are being finished.
# ## Solution:
# # Have a requirement that all of the tasks will be added before waiting happens.
# # Solution have a set of tasks, task has to remove itself from the running taskset when its done.
#     # Handle the case, when the task dies.





import asyncio
from typing import Callable, Coroutine, Optional, Union
from threading import Thread


class ThrottledTasksExecutor:
    """Executor which allows to run coroutines without hitting throttling limits.

    Usage example:
        async def task(name: str) -> str:
            print(f"Started greeting generation for {name}")
            await asyncio.sleep(1)
            print("Done greeting generation for {name}")
            return f"Hello, {name}!"

        def process_result(greeting: str) -> None:
            print("See what we got: {greeting}")

        with ThrottledTasksExecutor(delay_between_tasks=0.2) as executor:
            executor.run(generate_greeting("World"), callback=process_response)
            executor.run(generate_greeting("Universe"), callback=process_response)
    """

    def __init__(self, delay_between_tasks: float = 0.2, in_separate_process: bool = False):
        self.loop = asyncio.new_event_loop()
        self.running_tasks = set()  # TODO: Find the example with task removal from the set.
                                    # TODO: Would be nice to have awaitable future, instead of infinite loop.
        self.delay_between_tasks = 1
        self.in_separate_process = in_separate_process

        self.can_task_be_executed = asyncio.Condition(loop=self.loop)
        self.is_running = False

    def __enter__(self) -> "ThrottledTasksExecutor":
        self.start()
        return self

    def __exit__(self, *exc_info):
        self.wait_for_tasks_to_finish()
        self.stop()

    def start(self, in_separate_process: Optional[bool] = None):
        """Starts a thread (or a process), which executes coroutines provided to the ThrottledTasksExecutor"""

        if in_separate_process is None:
            in_separate_process = self.in_separate_process

        if in_separate_process:
            # TODO: investigate a way to start coroutines in a separate process:
            #   - https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
            raise NotImplemented("Running executor in a separate process is not supported yet")
        else:
            thread = Thread(target=self._run_event_loop, daemon=True)
            thread.start()

        self.is_running = True

        # Periodically emit permission to execute one task
        asyncio.run_coroutine_threadsafe(self._allow_task_execution(every=self.delay_between_tasks), self.loop)

    def stop(self):
        """Terminates a thread (or a process), which executes coroutines provided to the ThrottledTasksExecutor"""

        self.loop.stop()
        self.is_running = False

    def run(self, coroutine: Union[Coroutine, Callable], *args, callback: Optional[Callable] = None, **kwargs):
        """Executes coroutine in an executor thread, which makes sure not to hit throttling limits."""

        if callback is None:
            callback = lambda *args, **kwargs: None

        if not isinstance(coroutine, Coroutine):
            coroutine = coroutine(*args, **kwargs)
        if not isinstance(coroutine, Coroutine):
            raise ValueError("Can only execute coroutines, not coroutine provided")

        task = asyncio.run_coroutine_threadsafe(self._throttled_task(coroutine), self.loop)
        self.running_tasks.add(task)
        task.add_done_callback(self._mark_task_done(callback))

    def run_not_throttled(self, coroutine: Union[Coroutine, Callable], *args, callback: Optional[Callable] = None, **kwargs):
        """Executes coroutine in an executor event loop ignoring throttled tasks queue."""

        if callback is None:
            callback = lambda *args, **kwargs: None

        if not isinstance(coroutine, Coroutine):
            coroutine = coroutine(*args, **kwargs)
        if not isinstance(coroutine, Coroutine):
            raise ValueError("Can only execute coroutines, not coroutine provided")

        task = asyncio.run_coroutine_threadsafe(coroutine, self.loop)
        self.running_tasks.add(task)
        task.add_done_callback(self._mark_task_done(callback))

    def wait_for_tasks_to_finish(self):
        asyncio.run(self.async_wait_for_tasks_to_finish())

    async def async_wait_for_tasks_to_finish(self):
        while self.running_tasks:
            await asyncio.sleep(0.1)

    async def _allow_task_execution(self, every: float, count: int = 1):
        """Periodically emits event, which allows for `count` tasks to be executed.

        Params:
            every: float  - seconds how often a permission is given to `count` tasks to be executed.
            count: int    - how many tasks are allowed to be run at each event.
        """
        while self.is_running:
            # Does the first task has to wait `every` seconds to be started?
            async with self.can_task_be_executed:
                if count and count > 1:
                    self.can_task_be_executed.notify(n=count)
                else:
                    self.can_task_be_executed.notify_all()
            await asyncio.sleep(every)

    def _throttled_task(self, coroutine: Coroutine) -> Coroutine:
        """Decorator for coroutine to wait for permission before executing the coroutine."""
        async def throttled_task_wrapper(*args, **kwargs):
            async with self.can_task_be_executed:
                await self.can_task_be_executed.wait()
            return await coroutine
        return throttled_task_wrapper()

    def _mark_task_done(self, callback):
        """Decorator for callback to set the task as done after the callback is processed."""
        def task_done_wrapper(task):
            task_result = task.result()
            callback_result = callback(task_result)
            self.running_tasks.discard(task)
            return callback_result
        return task_done_wrapper

    def _run_event_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()



from unittest import TestCase
import logging

from gitmultirepoupdater.cli import parse_command_line_arguments


class CommandLineArgumentParsingTests(TestCase):
    def test_repositories_parsing(self):
        args = parse_command_line_arguments(["-r", "foo"])
        self.assertEqual(args.repos, ["foo"])

        args = parse_command_line_arguments(["-r", "foo", "-r", "bar", "-r", "spam"])
        self.assertEqual(args.repos, ["foo", "bar", "spam"])

        args = parse_command_line_arguments(["--repo", "foo", "--repos", "bar", "-r", "spam"])
        self.assertEqual(args.repos, ["foo", "bar", "spam"])

        args = parse_command_line_arguments(["--repos", "foo", "bar"])
        self.assertEqual(args.repos, ["foo"])

        args = parse_command_line_arguments([])
        self.assertEqual(args.repos, [])

    def test_clone_to_parsing(self):
        args = parse_command_line_arguments(["-r", "foo"])  # Default value
        self.assertEqual(args.clone_to, "/tmp/")

        args = parse_command_line_arguments(["-r", "foo", "-c", "my_dir"])
        self.assertEqual(args.clone_to, "my_dir")

        args = parse_command_line_arguments(["-r", "foo", "--clone-to", "my_another_dir"])
        self.assertEqual(args.clone_to, "my_another_dir")

    def test_verbose_parsing(self):
        args = parse_command_line_arguments([])
        self.assertFalse(args.verbose)
        self.assertEqual(logging.getLogger().level, logging.WARNING)

        args = parse_command_line_arguments(["-v"])
        self.assertTrue(args.verbose)
        self.assertEqual(logging.getLogger().level, logging.DEBUG)

        args = parse_command_line_arguments(["--verbose"])
        self.assertTrue(args.verbose)
        self.assertEqual(logging.getLogger().level, logging.DEBUG)

    def test_commands_parsing(self):
        args = parse_command_line_arguments(["ls"])
        self.assertEqual(args.commands, ["ls"])

        args = parse_command_line_arguments(["ls", "pwd"])
        self.assertEqual(args.commands, ["ls", "pwd"])

        args = parse_command_line_arguments(["-r", "repos.txt", "ls", "pwd"])
        self.assertEqual(args.commands, ["ls", "pwd"])



import argparse
import logging
import sys
from typing import List, Optional
from gitmultirepoupdater.data_types import CliArguments

from gitmultirepoupdater.utils.helpers import get_random_hex


def get_argument_parser():
    parser = argparse.ArgumentParser(
        description="Update multiple GitLab or GitHub repositories with a single command.",
        epilog="""Report bugs and request features at https://github.com/albertas/git-multi-repo-updater/issues""",
        add_help=False)

    parser.add_argument(
        "-r", "--repos", action="append", dest="repos", nargs="?", default=[],
        type=str, help="Repository url or Path to a file containing list of repository urls")

    parser.add_argument(
        "-c", "--clone-to", action="store", dest="clone_to", default="/tmp/", type=str, nargs="?",
        help="Path to directory which will be used to clone git repositories to (default is /tmp/)")

    parser.add_argument(
        "-m", "--message", "--commit-message", action="store", dest="commit_message",
        default=f"Git multi repo updater action #{get_random_hex()}", type=str, nargs="?",
        help="Message which will be used for commit message (and for branch name, PR title if they are not specified)")

    parser.add_argument(
        "-v", "--verbose", action=argparse.BooleanOptionalAction, dest="verbose",
        default=False, type=bool,
        help="Increase verbosity and show DEBUG logs")

    parser.add_argument(
        "-b", "--branch", action="store", dest="branch", type=str, nargs="?",
        help="Branch name which will be used to apply changes (it will be created if it does not exist)")

    parser.add_argument(
        "-t", "--target-branch", action="store", dest="target_branch", type=str, nargs="?",
        help="Target branch name which will be used to pull changes into")

    parser.add_argument(
        "-h", "--help", action="help", default=argparse.SUPPRESS,
        help="Show this message and exit.")

    parser.add_argument(
        "commands", type=str, nargs="*",
        help="Commands or scripts to execute in local Git repository")

    return parser


def parse_command_line_arguments(args: Optional[List[str]] = None) -> CliArguments:
    if args is None:
        args = sys.argv[1:]

    parser = get_argument_parser()
    parsed_args: CliArguments = parser.parse_args(args=args)  # type: ignore

    if parsed_args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    return parsed_args



from dataclasses import dataclass
from typing import Dict, List, Optional
from gitmultirepoupdater.constants import CloningStates, PullRequestStates, ModificationState


@dataclass
class CliArguments:
    repos: List[str]
    clone_to: str
    commands: List[str]
    commit_message: str
    verbose: bool
    branch: Optional[str]


@dataclass
class RepoState:
    args: CliArguments  # Parsed command line arguments 

    source_branch: str = ""  # Branch name from which a new branch for changes will be created
    branch: str = ""  # Branch name in which changes will be made and commited
    target_branch: str = ""  # Base branch into which PR changes will be pulled

    cloning_state: str = CloningStates.NOT_STARTED.value
    modification_state: str = ModificationState.NOT_STARTED.value
    pull_request_state: str = PullRequestStates.NOT_CREATED.value
    pull_request_status_code: Optional[int] = None
    pull_request_reason: Optional[str] = None

    name: str = ""  # Short human readable repo identifier
    owner: str = ""  # Owner of this repo
    url: str = ""  # Url used to clone the repository
    domain: str = ""  # Domain where the remote repository is hosted at (parsed from url)
    pull_request_url: str = ""  # Link to created pull request
    directory: str = ""  # Repository path in the file system

    stdout: str = ""  # Standard output from command execution
    stderr: str = ""  # Standard error output from command execution


@dataclass
class HttpRequestParams:
    url: str
    headers: Dict[str, str]
    data: Dict[str, str]



from typing import Dict
import git
from git.cmd import Git

from gitmultirepoupdater.utils.helpers import to_kebab_case
from gitmultirepoupdater.data_types import RepoState
from gitmultirepoupdater.utils.throttled_tasks_executor import ThrottledTasksExecutor


async def create_branch(repo: RepoState):
    new_branch_name = to_kebab_case(repo.args.commit_message)
    repo.branch = new_branch_name

    g = Git(repo.directory)
    g.execute(["git", "checkout", "-b", repo.branch])
    g.execute(["git", "pull", "origin", repo.branch])


def create_branch_for_each_repo(repos: Dict[str, RepoState], executor: ThrottledTasksExecutor) -> None:
    for repo in repos.values():
        executor.run_not_throttled(create_branch(repo))
    executor.wait_for_tasks_to_finish()



from typing import Dict
import git
from logging import getLogger
from gitmultirepoupdater.constants import ModificationState

from gitmultirepoupdater.data_types import RepoState
from gitmultirepoupdater.utils.throttled_tasks_executor import ThrottledTasksExecutor


async def commit_and_push_changes(repo: RepoState) -> None:
    g = git.Repo(repo.directory)

    if g.index.diff(None) or g.untracked_files:
        g.git.add(A=True)
        g.git.commit(m=repo.args.commit_message)

        g.git.push("--set-upstream", "origin", repo.branch)
        repo.modification_state = ModificationState.PUSHED_TO_REMOTE.value
    else:
        repo.modification_state = ModificationState.NO_FILES_CHANGED.value


def print_modified_repositories(repos: Dict[str, RepoState]):
    print()
    print("\033[1;34m|" + "Updated repositories".center(77, "-") + "|\033[0m")
    should_print_not_modified_repos = False
    print_repo_exceptions = False
    for repo in repos.values():
        if repo.modification_state == ModificationState.PUSHED_TO_REMOTE.value:
            print(f"\033[1;34m|\033[0m - {repo.url.ljust(73, ' ')} \033[1;34m|\033[0m")
        else:
            if repo.modification_state == ModificationState.GOT_EXCEPTION.value:
                print_repo_exceptions = True
            should_print_not_modified_repos = True

    if should_print_not_modified_repos:
        print("\033[1;34m|\033[0m" + "Did NOT modify these repositories:".center(77, "-") + "\033[1;34m|\033[0m")
        for repo in repos.values():
            if repo.cloning_state != ModificationState.PUSHED_TO_REMOTE.value:
                print(f"\033[1;34m|\033[0m - {(repo.url + ' ' + repo.modification_state).ljust(73, ' ')} \033[1;34m|\033[0m")

    if print_repo_exceptions:
        print("\033[1;34m|\033[0m" + "Exceptions:".center(77, "-") + "\033[1;34m|\033[0m")
        for repo in repos.values():
            if repo.cloning_state == ModificationState.GOT_EXCEPTION.value:
                print("\033[1;34m|\033[0m" + f" - {(repo.url + ' ' + repo.modification_state).ljust(73, ' ')}:" + "\033[1;34m|\033[0m")
                print(repo.stderr)

    print("\033[1;34m|" + "".center(77, "-") + "|\033[0m")


def commit_and_push_changes_for_each_repo(repos: Dict[str, RepoState], executor: ThrottledTasksExecutor) -> None:
    for repo in repos.values():
        executor.run(commit_and_push_changes(repo))
    executor.wait_for_tasks_to_finish()
    print_modified_repositories(repos)



from typing import List, Optional
from gitmultirepoupdater.actions.argument_parsing import parse_command_line_arguments
from gitmultirepoupdater.actions.get_repository_states import get_repository_states
from gitmultirepoupdater.actions.clone_repositories import clone_repositories
from gitmultirepoupdater.actions.create_branch import create_branch_for_each_repo
from gitmultirepoupdater.actions.run_command import run_command_for_each_repo
from gitmultirepoupdater.actions.commit_and_push_changes import commit_and_push_changes_for_each_repo 
from gitmultirepoupdater.actions.create_pull_request import create_pull_request_for_each_repo
from gitmultirepoupdater.utils.throttled_tasks_executor import ThrottledTasksExecutor


def main(args: Optional[List[str]] = None) -> None:
    cli_args = parse_command_line_arguments(args)
    repos = get_repository_states(cli_args)

    with ThrottledTasksExecutor(delay_between_tasks=0.1) as executor:
        clone_repositories(repos, executor)
        create_branch_for_each_repo(repos, executor)
        run_command_for_each_repo(repos, executor)
        commit_and_push_changes_for_each_repo(repos, executor)
        create_pull_request_for_each_repo(repos, executor)


if __name__ == "__main__":
    main()



from unittest import TestCase

from gitmultirepoupdater.cli import main


# class CommandLineArgumentParsingTests(TestCase):
#     def test_repositories_parsing(self):
#         args = main(["-r", "foo"])
#         self.assertEqual(args.repos, ["foo"])



import os
from urllib.parse import urlparse
from typing import List, Optional
from random import randint
from string import ascii_letters, digits

from git.cmd import Git

from gitmultirepoupdater.constants import ACCESS_TOKEN_VAR_NAMES
from gitmultirepoupdater.data_types import RepoState


def flatten_list(list_of_lists: Optional[List[List[str]]]) -> List[str]:
    """Concatenates lists into one list."""
    if not list_of_lists:
        return []
    return [elem for elem_list in list_of_lists for elem in elem_list]


def remove_suffix(value: str, suffix: str, case_insensitive: bool = True) -> str:
    """Removes suffix from a given string `value` case_insensitive by default."""
    if (
        value.endswith(suffix)
        or (case_insensitive and value.lower().endswith(suffix.lower()))
    ):
        return value[:-len(suffix)]
    return value


def get_random_hex() -> str:
    """Returns 8 digit hex code."""
    return hex(randint(2**31 + 1, 2**32))[2:]


def to_kebab_case(value: str) -> str:
    """Converts provided text `value` to a kebab-case name with max length of 100 for a branch name."""
    allowed_chars = ascii_letters + digits + " -/"
    value = value.replace(".", " ").replace(",", " ").replace("\\", " ").replace(":", " ").replace(";", " ")
    ascii_value = "".join([l for l in value if l in allowed_chars])
    return "-".join(ascii_value.lower().split())[:100]


def get_access_token(url: str) -> str:
    """Retrieves access token from environment vars suitable for provided URL."""
    parsed_url = urlparse(url)
    domain = parsed_url.netloc.split("@")[-1].lower()
    access_token_var_name = ACCESS_TOKEN_VAR_NAMES.get(domain, ACCESS_TOKEN_VAR_NAMES["DEFAULT"])
    return os.getenv(access_token_var_name, "")


def get_domain(url: str) -> str:
    """Returs domain name in lower case."""
    parsed_url = urlparse(url)
    return parsed_url.netloc.split("@")[-1].lower()


def get_repo_owner(url: str) -> str:
    """Parses repository owner from a valid Git url."""
    return url.rsplit("/", 2)[1]


def get_repo_name(url: str) -> str:
    """Parses repository name from a valid Git url."""
    return remove_suffix(url.split("/")[-1], ".git")


def get_default_branch(repo: RepoState):
    g = Git(repo.directory)
    default_branch_name: str = g.execute(["git", "rev-parse", "--abbrev-ref", "origin/HEAD"])  # type: ignore
    default_branch_name = default_branch_name.split("/", 1)[-1]  # removes `origin/` prefix from the result
    return default_branch_name
